#!usr/bin/env python3
"""
Tidal propagation cost function methodology from Zhang, Wright, et al., (2021)
Using fast marching algorithm and channel mask, determine optimal bathymetric
modification for improved tidal propagation in a hydrodynamic model.
"""
import numpy as np
import matplotlib.pyplot as plt
import os
import pandas as pd
import rasterio
import skfmm

#--------------------------------------
# Inputs
#--------------------------------------
# Binary channel mask from remotely-sensed imagery:
channels_path = 'path/to/channel_mask.tif'
# Mask delineating ocean or coastline, origin of tides:
coast_path = 'path/to/delineated/coastline_mask.tif'
# Unmodified DEM bathymetry/topography:
bathymetry_path = 'path/to/bathymetry.tif'
# Optional mask delineating acceptable carving region:
mask_path = 'path/to/optional/masked_area.tif'
# Folder path to save outputs
results_folder = 'path/to/results_folder'

sea_level = 0.3 # Assumed flat water surface elevation
min_cel = 1 # Minimum celerity value assigned to zero-depth cells

# Gauge indices (list of tuples)
gauges_inds = [(250, 700), (50, 1100)]
# NOTE: If all you have is gauge coordinates, consider the `coord2ind`
# function in the `dorado` python package (pip install pydorado)

# In this code, we will implement bathy modifications as a constant carving
# depth dZ everywhere inside a channel cell. This can be subbed out for more
# sophisticated methods. Here, we will test five non-zero dZ values:
carving_dZ = np.array([0, -0.5, -1, -2, -5, -10])
# NOTE: first element should be 0

#--------------------------------------
# Load data
#--------------------------------------
channels = rasterio.open(channels_path).read(1)
coast = rasterio.open(coast_path).read(1)
try:
    mask = rasterio.open(mask_path).read(1)
except:
    mask = np.ones_like(channels)
src = rasterio.open(bathymetry_path)
bathy = src.read(1).astype(float)
nodata = src.profile['nodata']
if nodata is not None:
    bathy[bathy==nodata] = np.nan
dx = src.profile['transform'][0] # Grid size
src.profile['dtype'] = 'float32'
src.profile['compress'] = 'deflate'

#--------------------------------------
# Apply method
#--------------------------------------
total_time = np.ones_like(carving_dZ)*np.nan
delta_mass = np.ones_like(carving_dZ)*np.nan

for n, k in enumerate(carving_dZ):
    # Modify the bathy
    modified_bathy = bathy.copy()
    modified_bathy[channels == 1] += k
    # Don't change in excluded areas
    modified_bathy[mask==0] = bathy[mask==0]

    # Get depth and celerity
    depth = sea_level - modified_bathy
    depth[modified_bathy >= sea_level] = 0.01

    celerity = np.sqrt(depth*9.81)
    celerity[modified_bathy >= sea_level] = min_cel

    # Compute travel time
    phi = (coast == 0) # Origin of tides
    travtime_xy = skfmm.travel_time(phi, speed=celerity, dx=dx, order=2)

    # Record performance data
    delta_mass[n] = np.sum(bathy - modified_bathy)*dx**2
    gaugetime = np.zeros(len(gauge_inds))
    for nn, ii in enumerate(gauge_inds):
        # Grab times at each gauge
        gaugetime[nn] = travtime_xy[gauge_inds[0],gauge_inds[1]]
    total_time[netname][n] = np.sum(gaugetime)

    # Save travel time geotiff
    outfile = os.path.join(results_folder, 'TravelTime_dZ%s.tif' % k)
    with rasterio.open(outfile, 'w', **src.profile) as dst:
        dst.write(travtime_xy.astype(np.float32), 1)

    # Save figure
    if n==0:
        # Initialize figure
        f, ax = plt.subplots(figsize=(5,5), dpi=300)
        im = ax.imshow(travtime_xy, cmap='cubehelix',
                       vmax=np.percentile(travtime_xy, 90))
        ax.set_title('Tidal Propagation Time, $t_{x,y}$')
        ax.set_facecolor('black')
        cbar = f.colorbar(im, ax=ax, fraction=0.04)
        ax.scatter(gauge_inds[0],gauge_inds[1],c='r',marker='o',s=1)
    else:
        im.set_data(travtime_xy) # Update with new data
        plt.draw()
    plt.savefig(outfile.replace('tif','png'), bbox_inches='tight')
plt.close()
# Normalize all by initial time
total_time /= total_time[0]
#--------------------------------------
# Save results
#--------------------------------------
df = pd.DataFrame()
df['Carving: dZ'] = carving_dZ
df['Mass: dM'] = delta_mass
df['Times: St'] = total_time
df.to_csv(os.path.join(results_folder,'PerformanceMetrics.csv'), index=False)

#--------------------------------------
# Plot Pareto front
#--------------------------------------
f, ax = plt.subplots(1, 1, figsize=(5,5), dpi=300)
ax.plot(delta_mass, total_time, color='k',
        marker='o', linewidth=1.5, markersize=7, zorder=0)
# Add carving depth label to scatter:
for i, txt in enumerate(carving_dZ):
    ax.annotate(str(abs(txt)), (delta_mass[i], total_time[i]),
                fontsize=4, color='w', ha="center", va="center",
                weight='extra bold', zorder=1)
ax.set_xlim([0, np.max(delta_mass)*1.1])
ax.set_ylim([np.min(total_time)*0.9, 1])
ax.set_xlabel('Mass Removed $[m^3]$')
ax.set_ylabel('Reduction in At-Gauge Travel Times $[-]$')
ax.set_facecolor('whitesmoke')
plt.savefig(os.path.join(results_folder, 'ParetoFront.png'),
            bbox_inches='tight')
plt.close()

